// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// getgravity.cc                                                               |
//                                                                             |
// copyright Walter Dehnen, 2002-2003                                          |
// e-mail:   walter.dehnen@astro.le.ac.uk                                      |
// address:  Department of Physics and Astronomy, University of Leicester      |
//           University Road, Leicester LE1 7RH, United Kingdom                |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// computes gravity (pot & acc) at sink position due to other sources.         |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// history:                                                                    |
//                                                                             |
// v 0.0    23/11/2002  WD created.                                            |
// v 0.1    04/02/2003  WD default falcON parameters automized                 |
// v 0.2    20/03/2003  WD gravity, action reporting                           |
// v 0.3    23/05/2003  WD automated NEMO history                              |
//-----------------------------------------------------------------------------+
#ifndef falcON_NEMO
#error You need NEMO to compile "src/mains/getgravity.cc"
#endif
#include <body.h>                                  // bodies etc..              
#include <falcON.h>                                // fakcIB                    
#include <public/nmio.h>                           // WDs C++ NEMO I/O          
#include <main.h>                                  // main & NEMO stuff         
//------------------------------------------------------------------------------
string defv[] = {
  "srce=???\n          input file: sources [m,x]       ",
  "sink=???\n          input file: sinks   [x]         ",
  "out=???\n           output file         [x,a,p]     ",
  "times=all\n         time range (for srce only)      ",
  "eps=0.05\n          softening length                ",
  "kernel="falcON_KERNEL_TEXT
  "\n                  softening kernel                ",
  "theta="falcON_THETA_TEXT
  "\n                  tolerance parameter at M=M_tot  ",
  "Ncrit="falcON_NCRIT_TEXT
  "\n                  max # bodies in un-split cells  ",
  falcON_DEFV, NULL };
//------------------------------------------------------------------------------
string usage="getgravity -- computes gravity at sink positions; using falcON";
//------------------------------------------------------------------------------
void nbdy::main()
{
  nemo_in  IN[2];
  IN[0].open(getparam("srce"));
  nemo_out OUT;
  unsigned NCRIT(getiparam("Ncrit"));
  io       READ[2],TOREAD[2] = {io::mx, io::x};
  const io WRITE = io::x | io::p | io::a;
  double   TIME;
  uint     NN[2];
  bodies   BODIES;
  falcON   FALCON(&BODIES,getrparam("eps"), getrparam("theta"),
		  kern_type(getiparam("kernel")));
  while(IN[0].is_present(nemo_io::snap)) {
    IN[1].open(getparam("sink"));
    if(BODIES.read_nemo_snapshots(IN,2,READ,NN,&TIME,TOREAD,getparam("times"))
       && READ[0] & TOREAD[0]
       && READ[1] & TOREAD[1]) {
      register real mass = BODIES.mass(0);
      for(register uint i=0; i!=NN[0]; ++i) {
	BODIES.flg(i).un_set(flag::ACTIVE);
	if(BODIES.mass(i) < mass) mass = BODIES.mass(i);
      }
      mass *= 1.e-10/real(NN[1]);
      for(register uint i=NN[0]; i!=BODIES.N_bodies(); ++i) {
	BODIES.flg(i).add(flag::ACTIVE);
	BODIES.mass(i) = mass;
      }
      FALCON.grow(NCRIT);
      FALCON.approximate_gravity();
      if(!OUT.is_open()) OUT.open(getparam("out"));
      BODIES.write_nemo_snapshot(OUT,&TIME,WRITE,NN[1],NN[0]);
    }
  }
}
